using System;
using Syncfusion.Windows.Forms ;
using GES.GridCellModels;
using System.Collections;
using Syncfusion.Windows.Forms.Grid;
using System.Reflection;
using System.Drawing;
using Syncfusion.Drawing;
using System.Windows.Forms;
using System.Collections.Specialized;
using System.ComponentModel;
using Syncfusion.ComponentModel;
using GES.Entities;
using Syncfusion.Windows.Forms.Tools.XPMenus;

namespace GES.Grid
{
	public delegate void GridDataUpdated();
	/// <summary>
	/// Summary description for GesDBGrid.
	/// </summary>
	public class DBGrid: Syncfusion.Windows.Forms.Grid.GridDataBoundGrid
	{
		public const string RichTextCell		= "GESRichText";
		public const string FormulaCell			= "GESFormulaCell";
		public const string StaticFormulaCell	= "StaticFormulaCell";
		public const string StaticCell			= "Static";
		public const string IntegerCell			= "GESIntegerCell";
		public const string IntegerStaticCell	= "IntegerStatic";
		public const string DecimalCell			= "DecimalCell";
		public const string NormalCell			= "GESTextBoxCell";
		public const string NoteCell			= "GESNoteCell";
		public const string SymbolCell			= "GESSymbolCell";
		public const string ItemCodeCell		= "GESItemCodeCell";
		public const string FormulaQuantityCell = "GESFormulaQuantityCell";
		public const string FormulaCurrencyCell = "GESFormulaCurrencyCell";
		public const string GroupCodeCell		= "GESGroupCodeCell";
		private int _gridCount;

		public const string CommandChangeCells	= "Change Cells";
		private ArrayList _columnInfo; // Store the column information.
		private HybridDictionary _fieldIndexes; // Stores the Indexes of the fields
		private ArrayList _dataSource; // Datasource Binded to the grid.
		private ArrayList _deletedRecords; // Strores the deleted records.

		StringCollection _richtextCols; 
		private bool _toBeSaved;
		private bool _readOnly;
		private Type _dataObjectType;
		private Hashtable formulaDependents;
		private ListDictionary richTextDependents;

		private Font _textFont;
		private Color _textColor;
		private string _roundUp;
		private string _currencyDecimals;
		bool _formulaExists;
		bool _cellEditing; 
		bool _colSettingsChanged;
		private bool _inSetColWidths;
		private bool isInitializing;
		
	
		// Events
		public event RecordAddedEventHandler RecordAdded;
		public event RecordEditedEventHandler RecordEdited;
		public event RecordDeletedEventHandler RecordDeleted;
		public event RecordDeletingEventHandler RecordDeleting;
		object oldRecord;


		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public ArrayList GesDataSource
		{
			get{return _dataSource;}
			set
			{
				_dataSource = value;
				this.DataSource = value;
			}
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public ArrayList ColumnInformation
		{
			get{return _columnInfo;}
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool ColumnSettingsChanged
		{
			get{return _colSettingsChanged;}
		}

		public DBGrid()
		{
			// Add the new cell Models
			this.Model.CellModels.Add(RichTextCell, new InPlaceRichTextCellModel(this.Model));
			this.Model.CellModels.Add(FormulaCell,new GES.GridCellModels.GesGridFormulaCellModel(this.Model));
			this.Model.CellModels.Add(IntegerCell, new GES.GridCellModels.GesIntegerCellModel(this.Model));
			this.Model.CellModels.Add(NormalCell, new GES.GridCellModels.GesTextBoxCellModel(this.Model));
			this.Model.CellModels.Add(NoteCell, new GES.GridCellModels.GesNoteCellModel(this.Model));
			this.Model.CellModels.Add(SymbolCell, new GES.GridCellModels.GesSymbolCellModel(this.Model,'+','%'));
			this.Model.CellModels.Add(ItemCodeCell, new GES.GridCellModels.GesItemCodeCellModel(this.Model));

			this.Model.Options.ActivateCurrentCellBehavior = GridCellActivateAction.None;
			this.Model.Options.EnterKeyBehavior = GridDirectionType.None;
			this.HScrollPixel = true;
			this.HorizontalThumbTrack  = true;
			this.VScrollPixel =true;
			this.VerticalThumbTrack = true;
			this.AllowSelection = GridSelectionFlags.AlphaBlend | GridSelectionFlags.Row | GridSelectionFlags.Multiple | GridSelectionFlags.Keyboard | GridSelectionFlags.Shift | GridSelectionFlags.Column;			
			this.Model.Options.SelectCellsMouseButtonsMask = MouseButtons.Left;
		}

		public void SetGridDetails(ArrayList dataSource, ArrayList columnInfo,Font normalTextFont,
			Color textColor, string roundUp,string currencyDecimals)
		{
			_dataObjectType = ((ArrayList)dataSource)[0].GetType(); 
			_currencyDecimals = currencyDecimals;
			_textFont = normalTextFont;
			_textColor = textColor;
			_columnInfo = columnInfo;
			_dataSource = dataSource;
			_roundUp = roundUp;
			_fieldIndexes = new HybridDictionary(true);
			_richtextCols = new StringCollection();
			InitializeGrid(dataSource);
			this.Model.ColWidthsChanged +=new GridRowColSizeChangedEventHandler(Model_ColWidthsChanged);
			this.Model.ColsMoved +=new GridRangeMovedEventHandler(Model_ColsMoved);
		}
		
		
		protected override void OnResize(EventArgs e)
		{
			base.OnResize (e);
			if(_columnInfo != null)
			{
				this.BeginUpdate();
				this.Model.CommandStack.Enabled =false;
				_inSetColWidths = true;
				SetColumnWidths();
				_inSetColWidths = false;
				this.Model.CommandStack.Enabled = true;
				this.EndUpdate();
				this.RefreshRange(this.ViewLayout.VisibleCellsRange);
			}
		}

		private void InitializeGrid(ArrayList dataSource)
		{
			if( _columnInfo == null)
				return;
			isInitializing = true;
			this.BeginUpdate();
			
			//Get the rows numbers
			this.Model.Options.NumberedRowHeaders = true;
			this.BaseStylesMap["Row Header"].StyleInfo.CellType = "Header";
			this.BaseStylesMap["Row Header"].StyleInfo.HorizontalAlignment = Syncfusion.Windows.Forms.Grid.GridHorizontalAlignment.Center;
	
			// Now resize all columns to fit contents.
			this.AllowResizeToFit = false;

			//turn off undo buffer so initial population cannot be undone
			this.Model.CommandStack.Enabled = false;
			

			#region Set Column and Row Header Color
			Color headerColorLeft	= SystemColors.Window;
			Color headerColorRight	= SystemColors.Control;
			//Set the color
			this.BaseStylesMap["Row Header"].StyleInfo.BackColor=Color.Gainsboro;
			this.BaseStylesMap["Row Header"].StyleInfo.Interior = new BrushInfo(GradientStyle.Vertical,headerColorLeft,headerColorRight );
			this.BaseStylesMap["Row Header"].StyleInfo.Font.Facename = "Arial";
			this.BaseStylesMap["Row Header"].StyleInfo.Font.Size  = 10F;
			this.BaseStylesMap["Row Header"].StyleInfo.Font.Bold  = true;
			
			this.BaseStylesMap["Column Header"].StyleInfo.BackColor=Color.Gainsboro;
			this.BaseStylesMap["Column Header"].StyleInfo.Interior =  new BrushInfo(GradientStyle.Vertical,headerColorLeft, headerColorRight );//Color.FromArgb(222, 236, 254),Color.FromArgb(147, 182, 232)
			this.BaseStylesMap["Column Header"].StyleInfo.Font.Facename = "Arial";
			this.BaseStylesMap["Column Header"].StyleInfo.Font.Size  = 10F;
			this.BaseStylesMap["Column Header"].StyleInfo.Font.Bold  = true;
			
			this.BaseStylesMap["Header"].StyleInfo.BackColor=Color.Gainsboro;
			this.BaseStylesMap["Header"].StyleInfo.Interior =  new BrushInfo(GradientStyle.Vertical,headerColorLeft, headerColorRight );
			this.BaseStylesMap["Header"].StyleInfo.Font.Facename = "Arial";
			this.BaseStylesMap["Header"].StyleInfo.Font.Size  = 10F;
			this.BaseStylesMap["Header"].StyleInfo.Font.Bold  = true;
			#endregion Set Column and Row Header Color

			#region Initialise Height
			//Default Row Height
			this.DefaultRowHeight =23;
				
			//Set the Row Header Size
			this.Model.RowHeights[0] = 40;
			this.Model.ColWidths[0] = 40;
			#endregion Initialise Height
			
			#region Bind Columns

			this.GridBoundColumns.Clear();
			for(int index = 0; index < _columnInfo.Count; index++)
			{
				GridBoundColumn gbc = new GridBoundColumn();
				gbc.MappingName = ((GesColumnStyle)_columnInfo[index]).FieldName ;
				_fieldIndexes.Add(gbc.MappingName,index);
				this.GridBoundColumns.Add(gbc);
			}
			
			#endregion
			
			
			SetColumnStyle();
			SetColumnWidths();
			
			this.Binder.DirectSaveCellInfo = true;
			this.DataSource = dataSource;
			ResizeToFitRichTextCells();
			//this.CommitChanges();
			//this.Refresh();
			this.Model.CommandStack.Enabled = true;
			this.EndUpdate(true);
			isInitializing = false;
		}

		public void ResizeToFitRichTextCells()
		{
			for(int i =0;i<_richtextCols.Count;i++)
			{
				int col = this.Binder.NameToColIndex(_richtextCols[i]);
				this.Model.RowHeights.ResizeToFit(GridRangeInfo.Cells(1,
					col,this.Model.RowCount,col),GridResizeToFitOptions.ResizeCoveredCells);//(int)_richtextCols[i]
			}
		}

		#region Set Column Widths 
		private void SetColumnWidths()
		{
			short totNumOfVisCols=0;
			short totNumOfRichText=0;
			short totNumOfNormal=0;

			for(int index = 0; index < _columnInfo.Count; index++)
			{
				GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[index];
				if(colStyle.Visible)
				{
					totNumOfVisCols++;
					if(colStyle.CellType == RichTextCell)
						totNumOfRichText++;
					else
						totNumOfNormal++;
				}
			}

			short usedWidth =0;
			
			for(int index = 0; index < _columnInfo.Count; index++)
			{
				GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[index];
				this.Model.HideCols[colStyle.FieldName] = !colStyle.Visible;
				this.Binder.InternalColumns[colStyle.FieldName].HeaderText = colStyle.HeaderText;

				if(colStyle.CellType != RichTextCell)
				{
					this.Model.ColWidths[colStyle.FieldName] = colStyle.CurrentWidth ;
					if(colStyle.Visible)
						usedWidth += colStyle.CurrentWidth ;
					
				}	

			}

			if(totNumOfRichText > 0)
			{
				short richTextWidth = (short)((this.Width - usedWidth - 40) / totNumOfRichText);

				for(int index = 0; index < _columnInfo.Count; index++)
				{
					GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[index];
					if(colStyle.CellType == RichTextCell)
						if(colStyle.CurrentWidth  < richTextWidth)
							this.Model.ColWidths[colStyle.FieldName] = richTextWidth;	
						else
							this.Model.ColWidths[colStyle.FieldName] = colStyle.CurrentWidth ;	
				}
			}

		}
		
		#endregion

		#region Set Column Styles
		private void SetColumnStyle()
		{
			_formulaExists = false;
			bool dependentFldExists = false;
			for(int index = 0; index < _columnInfo.Count; index++)
			{
				GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[index];
				GridStyleInfo styleInfo = this.Binder.InternalColumns[colStyle.FieldName].StyleInfo;
				switch (colStyle.CellType)
				{
					case FormulaCell:
					case FormulaQuantityCell:
					case FormulaCurrencyCell:
					case StaticFormulaCell:
						
						_formulaExists = true;
						if(formulaDependents == null)
							formulaDependents = new Hashtable();
					
						for(int i =0;i< ((ArrayList)this._dataSource).Count;i++)
						{
							colStyle.FormulaTags.Add(null);
						}
						// this.Binder.InternalColumns[colStyle.FieldName].StyleInfo.CellIdentity.ColIndex
						formulaDependents.Add(colStyle.FieldName,colStyle);
						
						styleInfo.HorizontalAlignment = GridHorizontalAlignment.Right;
						
						styleInfo.CellType = FormulaCell;
						styleInfo.WrapText = false;
						if(colStyle.CellType == FormulaCell)
							styleInfo.Format = "#,##0.00;(#,##0.00)";
						else if (colStyle.CellType == FormulaQuantityCell)
							styleInfo.Format = "#,##0." + _roundUp + ";(#,##0." + _roundUp + ")";
						else if (colStyle.CellType == FormulaCurrencyCell)
							styleInfo.Format = "#,##0." + _currencyDecimals + ";(#,##0." + _currencyDecimals + ")";
						else
						{
							styleInfo.Format = "#,##0." + _currencyDecimals + ";(#,##0." + _currencyDecimals + ")";
							styleInfo.ReadOnly = true;
							styleInfo.Enabled = false;
						}
						break;
					case RichTextCell:
						_richtextCols.Add(colStyle.FieldName);
						styleInfo.CellType = RichTextCell;
						styleInfo.AutoSize = true; //allow row height to change as you type
						styleInfo.AllowEnter = false; //allow cell to use the enter key
						styleInfo.BackColor = this.BackColor;
						if(colStyle.DependentField != null && colStyle.DependentField.Length > 0)
						{
							if(richTextDependents == null)
								richTextDependents = new ListDictionary();
							dependentFldExists = true;
							richTextDependents.Add(colStyle.FieldName,colStyle);
						}
						break;
					case IntegerCell:
						styleInfo.Format = "#,###;(#,###)";
						styleInfo.HorizontalAlignment = GridHorizontalAlignment.Right;
						styleInfo.StrictValueType =true;
						styleInfo.CellType =  IntegerCell;
						styleInfo.MaxLength = colStyle.MaxLength;
						break;
					case IntegerStaticCell:
						styleInfo.CellType = StaticCell;
						styleInfo.HorizontalAlignment = GridHorizontalAlignment.Right;
						styleInfo.Format = "#,###;(#,###)";
						styleInfo.ReadOnly = true;
						styleInfo.Enabled = false;
						break;
					case DecimalCell:
						this.Binder.InternalColumns[colStyle.FieldName].StyleInfo.Format = "#,##0.00;(#,##0.00)";
						this.Binder.InternalColumns[colStyle.FieldName].StyleInfo.HorizontalAlignment = GridHorizontalAlignment.Right;
						//"#,###.00;(#,###.00)" 
						break;
					case StaticCell:
						styleInfo.CellType = StaticCell;
						styleInfo.ReadOnly = true;
						styleInfo.Enabled = false;	
						break;

					case NormalCell :
						styleInfo.CellType = NormalCell;
						styleInfo.WrapText =false;
						styleInfo.AllowEnter = false; //allow cell to use the enter key
						styleInfo.MaxLength = colStyle.MaxLength;
						//styleInfo.CharacterCasing = CharacterCasing.Upper;
						break;
					case NoteCell :
						this.Binder.InternalColumns[colStyle.FieldName].StyleInfo.CellType =  NoteCell;
						break;
					case SymbolCell :
						StringCollection items = new StringCollection();
						items.Add(string.Empty);
						items.Add("+");
						items.Add("%");
						styleInfo.CellType = "ComboBox";
						styleInfo.ChoiceList = items;
						styleInfo.ShowButtons = GridShowButtons.ShowCurrentCell;
						styleInfo.DropDownStyle = GridDropDownStyle.Exclusive;
						styleInfo.HorizontalAlignment = Syncfusion.Windows.Forms.Grid.GridHorizontalAlignment.Center;
						styleInfo.MaxLength = 1;
						styleInfo.AllowEnter = false;
						break;
					case ItemCodeCell :
						styleInfo.CellType = ItemCodeCell;
						styleInfo.WrapText = false;
						styleInfo.AllowEnter = false;
						styleInfo.MaxLength = colStyle.MaxLength;
						break;
					case GroupCodeCell :
						StringCollection Codeitems = new StringCollection();
						Codeitems.Add("All");
						Codeitems.Add("Q");Codeitems.Add("Q2");	Codeitems.Add("Q3");
						Codeitems.Add("Q4");Codeitems.Add("R");Codeitems.Add("R1");
						Codeitems.Add("R2");Codeitems.Add("R3");
						Codeitems.Add("R4");
						styleInfo.CellType = "ComboBox";
						styleInfo.ChoiceList = Codeitems;
						styleInfo.ShowButtons = GridShowButtons.ShowCurrentCell;
						styleInfo.DropDownStyle = GridDropDownStyle.Exclusive;
						styleInfo.HorizontalAlignment = Syncfusion.Windows.Forms.Grid.GridHorizontalAlignment.Left;
						styleInfo.MaxLength = 3;
						styleInfo.AllowEnter = false;
						break;
				}
				styleInfo.Font = new GridFontInfo(_textFont);
				styleInfo.TextColor = _textColor;
				if(_readOnly)
					styleInfo.ReadOnly = true;
			}
			

			if(_formulaExists || dependentFldExists)
			{
				if(_formulaExists)
					this.Model.QueryCellInfo+=new Syncfusion.Windows.Forms.Grid.GridQueryCellInfoEventHandler(Model_QueryCellInfo);
				this.Model.SaveCellInfo	+=new Syncfusion.Windows.Forms.Grid.GridSaveCellInfoEventHandler(Model_SaveCellInfo);
			}
			
		}

		#endregion

		#region QueryCellInfo & SaveCellInfo
		private void Model_QueryCellInfo(object sender, Syncfusion.Windows.Forms.Grid.GridQueryCellInfoEventArgs e)
		{
			if(_readOnly && e.RowIndex > 0 && e.ColIndex >0 )
				e.Style.ReadOnly = true;

			
			if(e.ColIndex > 0 && e.RowIndex > 0 && e.RowIndex <= ((ArrayList)this.DataSource).Count)
			{
				string fldName = this.Binder.InternalColumns[e.ColIndex-1].MappingName;
				if(_subTotalRows != null && _subTotalRows.Contains(e.RowIndex) && fldName != _subTotDescFld)
					e.Style.ReadOnly  = true;
				
				if(formulaDependents.ContainsKey(fldName))
				{
					// Assign the formula tag
					e.Style.FormulaTag = ((GesColumnStyle)formulaDependents[fldName]).FormulaTags[e.RowIndex - 1] as GridFormulaTag;	
				
					// Change the Text color to red incase Formula is entered
					double dummy;
					if(!double.TryParse(e.Style.Text,System.Globalization.NumberStyles.Any,null,out dummy))
						e.Style.TextColor = Color.Red;
					else
						e.Style.TextColor = this.ForeColor;
					e.Handled = true;
				}
			}
		}
		
		private void Model_SaveCellInfo(object sender, Syncfusion.Windows.Forms.Grid.GridSaveCellInfoEventArgs e)
		{
			if(e.RowIndex > 0 && e.RowIndex <= ((ArrayList)this.DataSource).Count && e.ColIndex > 0)
			{
				string fldName = this.Binder.InternalColumns[e.ColIndex-1].MappingName;
				if(formulaDependents != null && formulaDependents.ContainsKey(fldName))
				{					
					// Save the formual tag 
					bool inInit = false;
					if(((GesColumnStyle)formulaDependents[fldName]).FormulaTags[e.RowIndex-1] == null)
						inInit = true;
					
					((GesColumnStyle)formulaDependents[fldName]).FormulaTags[e.RowIndex-1] = e.Style.FormulaTag;
				
					double computedValue;
					// assing the computed value to the dependent field
					if(e.Style.FormulaTag == null || 
						!double.TryParse(e.Style.FormulaTag.Text,System.Globalization.NumberStyles.Any,null,out computedValue))
					{
						if(e.Style.Text == null ||  
							!double.TryParse(e.Style.Text,System.Globalization.NumberStyles.Any,null,out computedValue))
							computedValue = 0D;
					}
					
					object oldRecord = null;
					if(e.RowIndex != this.CurrentCell.RowIndex && RecordEdited != null)
					{
						if(inInit)
						{
							// Check if the value has changed
							// Incase of application level formula the value can change coz of update in another PL
							double oldValue = (double)_dataObjectType.InvokeMember(
								((GesColumnStyle)formulaDependents[fldName]).DependentField,
								BindingFlags.Default | BindingFlags.GetProperty, null,
								((ArrayList)this.DataSource)[e.RowIndex-1], null);
						
							if(oldValue != computedValue)
								inInit = false;
						}

						if(!inInit)
						{
							oldRecord = ((ICloneable)this._dataSource[e.RowIndex -1]).Clone();
						}
					}

					_dataObjectType.InvokeMember(((GesColumnStyle)formulaDependents[fldName]).DependentField,
						BindingFlags.Default | BindingFlags.SetProperty, null,
						((ArrayList)this.DataSource)[e.RowIndex-1],new object[] {computedValue});

					
					if(!inInit && e.RowIndex != this.CurrentCell.RowIndex)
					{
						if(RecordEdited != null)
							RecordEdited(this,new RecordEditedEventArgs(oldRecord,e.RowIndex-1,fldName));
					
						((GesGridRecordEntity)this._dataSource[e.RowIndex-1]).SetRowStateModified();
					}
				
					e.Handled = true;
				}
				else if (richTextDependents != null && richTextDependents.Contains(fldName))
				{
					_dataObjectType.InvokeMember(((GesColumnStyle)richTextDependents[fldName]).DependentField,
						BindingFlags.Default | BindingFlags.SetProperty, null,
						((ArrayList)this.DataSource)[e.RowIndex-1],new object[] {e.Style.FormattedText});
				}
			}
		}
		#endregion

		#region ADD / INSERT ITEM
		
		private void SetRecordRowNumber(object itemData, int row)
		{
			((GesGridRecordEntity)itemData).Row = row;
		}

		public void AddItem(object itemData)
		{
			SetRecordRowNumber(itemData,_dataSource.Count + 1);
			RaiseRecordAddedEvent(itemData);

			this._dataSource.Add(itemData);
			if(_formulaExists && this._dataSource != null)
			{
				IDictionaryEnumerator enumerator = formulaDependents.GetEnumerator();
				while (enumerator.MoveNext())
				{
					GesColumnStyle colStyle = enumerator.Value as GesColumnStyle;
					colStyle.FormulaTags.Add(null);
				}
			}
			//this.BeginUpdate();
			//this.RefreshRange(GridRangeInfo.Cells(this.Model.RowCount,0,this.Model.RowCount+1,this.Model.ColCount));
			//this.EndUpdate(true);
		}

		public void SetNewRowNumbers(int startIndex, int newRowStart)
		{
			int cnt =0;
			for(int i=startIndex;i < _dataSource.Count;i++)
			{	
				SetRecordRowNumber(_dataSource[i],newRowStart + cnt);
				cnt++;
			}
		}

		/// <summary>
		/// Inserts an item into the Grid Data Source 
		/// </summary>
		/// <param name="index">Index in the Arraylist (DataSource)</param>
		/// <param name="itemData">Item to be inserted</param>
		/// <param name="refreshRowNumber">True if Row Numbers have to be updated</param>
		public void InsertItem(int index, object itemData, bool refreshRowNumber)
		{	
			if(_formulaExists && this._dataSource != null)
			{
				string[] fieldNames = new string[formulaDependents.Count];
				int cnt = 0;
				IDictionaryEnumerator enumerator = formulaDependents.GetEnumerator();
				while (enumerator.MoveNext())
				{
					GesColumnStyle colStyle = enumerator.Value as GesColumnStyle;
					fieldNames[cnt] = colStyle.FieldName;
					cnt++;
					
					colStyle.FormulaTags.Insert(index,null);
					for(int i =0 ; i< colStyle.FormulaTags.Count ;i++)
						colStyle.FormulaTags[i] = null;
				}
				// Refresh Range in formula
				// Row Number starts from 1 - Index refers to index in arraylist 
				// Add 1 to index to get the row number
				AdjustReferencesForRangeChange(true,index + 1,index + 1,fieldNames);
			}

			SetRecordRowNumber(itemData, index + 1);
			RaiseRecordAddedEvent(itemData);
			if(refreshRowNumber)
			{
				this.BeginUpdate();
				this.Binder.SuspendBinding();
				this.Model.CommandStack.Enabled = false;
			}

			this._dataSource.Insert(index,itemData);
			
			if(refreshRowNumber)
			{
				SetNewRowNumbers(index + 1,index + 2);
				this.Model.CommandStack.Enabled = true;
				this.Binder.ResumeBinding();
				this.Model.ResetVolatileData();
				this.EndUpdate();
				this.ResizeToFitRichTextCells();
				this.RefreshRange(this.ViewLayout.VisibleCellsRange);
				_toBeSaved = true;
			}
		}

		//		public void InsertItem(int index, object itemData)
		//		{
		//			this.InsertItem(index,itemData,true);
		//			_toBeSaved =true;
		//		}

		public void InsertRows(int index, int count)
		{
			this.BeginUpdate();
			this.Binder.SuspendBinding();
			this.Model.CommandStack.Enabled = false;
			try
			{
				System.Reflection.ConstructorInfo oConstructor = _dataObjectType.GetConstructor(Type.EmptyTypes);
				for(int i =0 ; i < count; i++)
				{
					this.InsertItem(index + i, (oConstructor.Invoke(null)),false);
				}
				this.SetNewRowNumbers(index, index + 1);
				_toBeSaved = true;
				RefreshSubtotalRowNumbers(index,true,count);
			}
			catch
			{}
			finally
			{
				this.Model.CommandStack.Enabled = true;
				this.Binder.ResumeBinding();
				//this.Selections.Clear(false);
				this.Model.ResetVolatileData();
				this.CurrentCell.MoveTo(this.CurrentCell.RowIndex + count,this.CurrentCell.ColIndex); 
				this.CurrentCell.MoveTo(this.CurrentCell.RowIndex - count,this.CurrentCell.ColIndex); 	
				this.ResizeToFitRichTextCells();
				this.EndUpdate();
				this.RefreshRange(this.ViewLayout.VisibleCellsRange);
			
			}
		}
		
		public void InsertRows()
		{
			using(frmInsertRow frmInsert = new frmInsertRow(this))
			{
				frmInsert.ShowDialog();
			}
		}
		#endregion

		GridFindReplaceDialogSink findReplaceDialogSink;

		public GridFindReplaceDialogSink GridFindReplaceDialogSink
		{
			get
			{
				if (findReplaceDialogSink == null)
				{
					findReplaceDialogSink = new GridFindReplaceDialogSink(this);
				}
				return findReplaceDialogSink;
			}
		}

		#region REMOVE / Delete Rows
		//		public void RemoveItem(object obj)
		//		{
		//			int index = this._dataSource.IndexOf(obj);
		//			if(index == -1)
		//				return;
		//			RemoveItemAt(index);
		//		}

		protected override void OnRowsDeleting(GridRowRangeEventArgs e)
		{
			this.Binder.DirectSaveCellInfo = true;
			
			if(RecordDeleting != null)
			{
				CancelEventArgs args = new CancelEventArgs();
				RecordDeleting(this,args);
				if(args.Cancel)
				{
					e.Cancel = true;
					return;
				}
			}
			
			GridRangeInfoList rangeList = this.GetSelectedRows();
 
			if(rangeList.Count > 0)
			{
				ArrayList a = new ArrayList();
				foreach(GridRangeInfo range in rangeList)
				{
					for(int row = range.Top; row <= range.Bottom; ++row)
					{
						int pos = this.Binder.RowIndexToPosition(row);
						a.Add(pos);
					}
				}

				a.Sort();

				if(MessageBox.Show("Are you sure you want to delete these " + a.Count + " record(s)","Delete",
					MessageBoxButtons.YesNo,MessageBoxIcon.Question) == DialogResult.Yes)
				{
					this.Selections.Clear(false);
					this.Model.CommandStack.Enabled = false;
									
					if(_subTotalRows!= null && _subTotalRows.Count > 0)
					{
						int oldStartRow,oldEndRow;
						string dispText;
						bool deletedSubTotal = false;
						for(int j = 0;j< _subTotalRows.Count;j++)
						{
							if(!a.Contains((int)_subTotalRows[j]))
							{
								GetSubtotalStartEndRow((int)_subTotalRows[j] ,out oldStartRow,out oldEndRow,out dispText);
								int i =0;
								for(i = oldStartRow-1; i< oldEndRow; i++)
								{
									if(!a.Contains(i))
										break;
								}
								if(i == oldEndRow)
								{
									a.Add((int)_subTotalRows[j]-1);
									deletedSubTotal = true;
								}
							}
						}
						if(deletedSubTotal)
						{
							MessageBox.Show("Subtotal row is also deleted as all reffered rows are deleted.","Deleting Subtotal Row",MessageBoxButtons.OK,MessageBoxIcon.Information);
							a.Sort();
						}
					}
					//					for(int r=0;r < a.Count;r++)
					//						if ((int)a[r] + 1 < (int)((ArrayList)_subTotalRows)[1])
					//						{
					//							MessageBox.Show("This row(s) is referenced by SubTotal , delete that row also");
					//							e.Cancel = true;
					//							return;
					//						}

					//					for(int stCount;stCount < _subTotalRows.Count;stCount++)
					//						if (a[a.Count-1] < _subTotalRows[stCount])
					//							MessageBox.Show("This row(s) is referenced by SubTotal , delete that row also");
					
 
					
					this.BeginUpdate();
					this.Binder.SuspendBinding();

					for(int i = a.Count-1; i >= 0; i--)
						this.RemoveRange((int)a[i]);
					
					
					this.SetNewRowNumbers((int)a[0] ,(int)a[0] + 1);

					for(int i =0 ; i < a.Count ;i++)
					{
						try
						{
							System.Reflection.ConstructorInfo oConstructor = _dataObjectType.GetConstructor(Type.EmptyTypes);
							this.AddItem(oConstructor.Invoke(null));
						}
						catch
						{}
					}

					this.Binder.ResumeBinding();
					this.Model.ResetVolatileData();
					_toBeSaved =true;
					this.RefreshRange(this.ViewLayout.VisibleCellsRange);
					this.Model.CommandStack.Enabled = true;
					this.CurrentCell.MoveTo(this.CurrentCell.RowIndex-1,this.CurrentCell.ColIndex); 
					this.CurrentCell.MoveTo(this.CurrentCell.RowIndex+1,this.CurrentCell.ColIndex);
					this.EndUpdate();
				}
			}
			e.Cancel = true;
			base.OnRowsDeleting (e);
			this.ResizeToFitRichTextCells();
			this.RefreshRange(this.ViewLayout.VisibleCellsRange);
		}


		public void RemoveRange(int index)//, bool deleteFromDatasource 
		{
			this.RemoveRange(index,true);
		}

		/// <summary>
		/// Override the base function 
		/// To delete the current row if a delete key is pressed
		/// </summary>
		/// <param name="e"></param>
		protected override void HandleBoundCurrentCellKeyDown(KeyEventArgs e)
		{
			if(e.KeyCode == Keys.Delete && !this.CurrentCell.IsEditing)
			{
				// if the row is not selected 
				// Select the Row
				GridRangeInfoList rangeList = this.GetSelectedRows();
				if(rangeList.Count == 0) // Full row is not selected
				{
					this.Selections.Clear();
					this.Selections.Add(GridRangeInfo.Row(this.CurrentCell.RowIndex));
				}
			}
			base.HandleBoundCurrentCellKeyDown (e);
		}

		public void RemoveRange(int index, bool addToDeletedRecords)//, bool deleteFromDatasource 
		{
			if(_formulaExists && this._dataSource != null)
			{
				string[] fieldNames = new string[formulaDependents.Count];
				int cnt = 0;
				IDictionaryEnumerator enumerator = formulaDependents.GetEnumerator();
				while (enumerator.MoveNext())
				{
					GesColumnStyle colStyle = enumerator.Value as GesColumnStyle;
					fieldNames[cnt] = colStyle.FieldName;
					cnt++;
					
					//colStyle.FormulaTags.RemoveRange(index,count);
					colStyle.FormulaTags.RemoveAt(index);
					for(int i =0 ; i< colStyle.FormulaTags.Count ;i++)
						colStyle.FormulaTags[i] = null;
				}
				
				// Refresh Range in formula
				// Row Number starts from 1 - Index refers to index in arraylist 
				// Add 1 to index to get the row number
				AdjustReferencesForRangeChange(false,index + 1,index + 1,fieldNames);
			}

			if(this._deletedRecords == null && addToDeletedRecords)
				_deletedRecords = new ArrayList();

			//			ArrayList delRecords = (this._dataSource.GetRange(index,1));
						
			Type _enumType = _dataObjectType.GetProperty("RowStatus").PropertyType;
			 
			//			for(int i =0; i< delRecords.Count ; i++)
			//			{	
			// Set the Row Status to deleted.
			
			if(	addToDeletedRecords  && !Enum.Equals(_dataObjectType.InvokeMember("RowStatus",
				BindingFlags.Default | BindingFlags.GetProperty,
				null,_dataSource[index],null) ,  Enum.ToObject(_enumType,2)))
			{
				
				_dataObjectType.InvokeMember("RowStatus",
					BindingFlags.Default | BindingFlags.SetProperty,
					null,_dataSource[index],new object[]{Enum.ToObject(_enumType,4)});

				_deletedRecords.Add(_dataSource[index]);
			}

			if(_subTotalRows != null &&  _subTotalRows.Count > 0 &&
				(bool)_dataObjectType.InvokeMember(_subTotFlagFld,
				BindingFlags.Default | BindingFlags.GetProperty,
				null,_dataSource[index],null) )
			{
				this.DeleteSubtotal(index);				
			}
			else
			{
				if(RecordDeleted != null && addToDeletedRecords)
				{
					RecordDeletedEventArgs e = new RecordDeletedEventArgs(_dataSource[index]);
					RecordDeleted(this,e);
				}
			}

			RefreshSubtotalRowNumbers(index + 1, false,1);	
			
			_dataSource.RemoveAt(index);
		}
	
		//		public void RemoveItemAt(int index)
		//		{
		//			this.Model.Rows.RemoveRange(index + 1,index + 1);
		//		}
		
		#endregion

		#region Adjust Range 
		private void AdjustReferencesForRangeChange(bool isInsert, int start, int end, params string[] fieldNames)
		{
			int offSet = isInsert ? end - start + 1 
				: start - end - 1;
			string oldFormula;
			
			for(int index = 0; index < _dataSource.Count; index++)
			{
				for(int i =0;i < fieldNames.Length ;i++)
				{
					oldFormula = (string)_dataObjectType.InvokeMember(fieldNames[i],
						BindingFlags.Default | BindingFlags.GetProperty, null,
						((ArrayList)this.DataSource)[index],new object[]{});

					if(oldFormula != null && oldFormula.Length > 0)
					{
						oldFormula = oldFormula.ToUpper();
		
						if(oldFormula.IndexOf("SUM(") != -1 || oldFormula.IndexOf("MAX(") != -1 
							|| oldFormula.IndexOf("MIN(") != -1 || oldFormula.IndexOf("AVG(") != -1
							|| oldFormula.IndexOf("#") != -1 || oldFormula.IndexOf("PRODUCT(") != -1
							|| oldFormula.IndexOf("SUMSQ(") !=-1)
						{
							string newFormula = AdjustReferences(oldFormula, offSet, start);
							
							if(string.Compare(newFormula,oldFormula,true)!= 0)
								_dataObjectType.InvokeMember(fieldNames[i],
									BindingFlags.Default | BindingFlags.SetProperty, null,
									((ArrayList)this.DataSource)[index],new object[] {newFormula});

						}
					}
				}
			}
		}

		private string AdjustReferences(string origText, int rowOffset, int rowCut)
		{
			string s = "";
			int pos = 0;
			int len = origText.Length;

			while (pos < len)
			{
				if(rowOffset !=0 && pos + 4 < len && 
					(origText.Substring(pos,4)== "SUM(" || origText.Substring(pos,4)== "AVG(" || 
					origText.Substring(pos,4)== "MIN(" || origText.Substring(pos,4)== "MAX(" ))
				{
					s += RefreshGESRange(ref pos,origText,4,rowOffset,rowCut);
				}
					// Added Nous
					// To refresh range incase of a MaxA() / MinA()
				else if(rowOffset !=0 && pos + 8 < len && origText.Substring(pos,8)== "PRODUCT(" )
				{
					s += RefreshGESRange(ref pos,origText,8,rowOffset,rowCut);
				}
				else if(rowOffset !=0 && pos + 6 < len && origText.Substring(pos,6)== "SUMSQ(" )
				{
					s += RefreshGESRange(ref pos,origText,6,rowOffset,rowCut);
				}
				else if(rowOffset !=0 && origText[pos] == '#' && char.IsDigit(origText,pos+1))
				{
					pos++;
					int loopStart = pos;
					while(pos < len && char.IsDigit(origText, pos))
						pos++;
					
					int row = int.Parse(origText.Substring(loopStart, pos - loopStart));
					if(row >= rowCut)
						row = Math.Max(row + rowOffset, this.Model.Rows.HeaderCount + 1);
					s += '#' + row.ToString();
				}
				
				else
				{
					s += origText[pos];
					pos++;
				}
			}
			//	Console.WriteLine("out[" + s + "]");
			return s;
		}

		private string RefreshGESRange(ref int pos,string upperText,int funcLen,int rowOffset, int rowCut)
		{
			string funcName = upperText.Substring(pos,funcLen);
			
			pos+=funcLen;
					
			int loopStart = pos;
			while(pos < upperText.Length && upperText[pos] != ')')
				pos++;
			string cellRef = upperText.Substring(loopStart, pos - loopStart);

			string[] dfv = null;
			if(cellRef.IndexOf(',') != -1)
				dfv= cellRef.Split(',');
			else if(cellRef.IndexOf(':') !=-1)
				dfv = cellRef.Split(':');
				
			string s2  = string.Empty;

			foreach(string sRow in dfv)
			{
				if(char.IsDigit(sRow[0]))
				{
					int row = int.Parse(sRow);
					if(row >= rowCut)
						row = Math.Max(row + rowOffset, this.Model.Rows.HeaderCount + 1);
					s2 +=  row.ToString() + ',';
				}
			}

			if(s2.Length > 0)
				s2 = s2.Substring(0,s2.Length -1);
			pos++;
			return funcName + s2 + ')';
		}

		public int RowIndex(string s)
		{
			int i = 0;
			
			while (i < s.Length && char.IsLetter(s[i]))
				i++;
			if(i < s.Length)
				return int.Parse(s.Substring(i));
			
			throw new ArgumentException("Bad Index");
		}

		#endregion

		public bool ReadOnly
		{
			get{return _readOnly;}
			set
			{
				_readOnly = value;
				this.EnableEdit = !value;
			}
		}

		protected override void OnPrepareViewStyleInfo(GridPrepareViewStyleInfoEventArgs e)
		{
			base.OnPrepareViewStyleInfo (e);
			if(_readOnly && e.RowIndex > 0 && e.ColIndex > 0)
				e.Style.BackColor = Color.Gainsboro;// SystemColors.Control;
			else if(_subTotalRows != null &&  _subTotalRows.Contains(e.RowIndex))
				e.Style.BackColor = Color.Gainsboro;
			//			else if(!_readOnly && e.RowIndex > 0 && e.ColIndex > 0 && e.RowIndex%2 ==0)
			//				e.Style.BackColor = Color.Snow;
			//			System.Diagnostics.Debug.WriteLine("Prepare");
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public ArrayList DeletedRecords
		{
			get{return _deletedRecords;}
		}

		#region SHOW / HIDE COLUMNS
		public void ShowColumns(params string[] fieldNames)
		{
			ShowHideColumns(true,fieldNames);
		}
		
		public void HideColumns(params string[] fieldNames)
		{
			ShowHideColumns(false,fieldNames);
		}

		private void ShowHideColumns(bool visible ,params string[] fieldNames)
		{
			if(fieldNames == null || fieldNames.Length == 0)
				throw new ArgumentNullException("fieldNames");
			
			bool columnVisibilityChanged = false;
			for(int i=0;i< fieldNames.Length; i++)
			{
				if(_fieldIndexes.Contains(fieldNames[i]))
				{
					GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[(int)_fieldIndexes[fieldNames[i]]];
					if(colStyle.Visible != visible)
					{
						colStyle.Visible = visible;
						columnVisibilityChanged =true;
					}
				}	
			}

			if(columnVisibilityChanged)
			{
				this.BeginUpdate();
				this.SetColumnWidths();
				this.EndUpdate(true);
				this.Refresh ();
				_colSettingsChanged = true;
			}
		}
		#endregion

		#region Raising Record Edited event for Status Bar Formula

		protected override void OnCurrentCellEditingComplete(EventArgs e)
		{
			base.OnCurrentCellEditingComplete (e);
			_cellEditing = false;
		}

		protected override void OnCurrentCellStartEditing(System.ComponentModel.CancelEventArgs e)
		{
			base.OnCurrentCellStartEditing (e);
			_cellEditing = true;

		}

		protected override void OnCurrentCellAcceptedChanges(System.ComponentModel.CancelEventArgs e)
		{
				
			this.Binder.EndEdit();
			base.OnCurrentCellAcceptedChanges(e);
			
			

			// Call the Record Edited event
			if(!isInitializing && this.RecordEdited != null && _cellEditing)
			{
				RecordEditedEventArgs args = new 
					RecordEditedEventArgs(oldRecord,this.CurrentCell.RowIndex-1,
					this.Binder.InternalColumns[this.CurrentCell.ColIndex -1].MappingName);
				RecordEdited(this,args);
			}			

			// Set the Row State to Modified 
			((GesGridRecordEntity)this._dataSource[this.CurrentCell.RowIndex -1]).SetRowStateModified();
			_toBeSaved =true;

			this.RecalculateRow(this.CurrentCell.RowIndex);
			this.RefreshRange(GridRangeInfo.Cells(this.CurrentCell.RowIndex ,0 ,this.CurrentCell.RowIndex,this.ViewLayout.LastVisibleCol));
			//This will tell the gridcount to move when CNTRL+END is invoked.
			if (this.GridCount < this.CurrentCell.RowIndex)
				this.GridCount = this.CurrentCell.RowIndex;
		}

		public void RecalculateRow(int rowIndex)
		{
			if(formulaDependents != null)
			{
				GesGridFormulaCellModel cellModel = (GesGridFormulaCellModel)this.Model.CellModels[FormulaCell];			 
				IDictionaryEnumerator myEnumerator = formulaDependents.GetEnumerator();
				while(myEnumerator.MoveNext())
				{
					GesColumnStyle colStyle = (GesColumnStyle)myEnumerator.Value;
					int colIndex = this.Binder.NameToColIndex(colStyle.FieldName);
					if(colIndex != this.CurrentCell.ColIndex)
						cellModel.RefreshCells(rowIndex,colIndex);
				}
			}
		}
		protected override void OnCurrentCellChanging(System.ComponentModel.CancelEventArgs e)
		{
			base.OnCurrentCellChanging (e);
			if(!isInitializing && this.RecordEdited != null && _cellEditing)
			{
				// get the old record before saving the changes
				oldRecord = ((ICloneable)((ArrayList)this.DataSource)[this.CurrentCell.RowIndex -1]).Clone(); 
			}
		}

		#endregion

		#region Save Flag 
		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool ToBeSaved
		{
			set{_toBeSaved = value;}
			get{return _toBeSaved;}
		}

		public void AcceptChanges()
		{
			if(this.CurrentCell.IsEditing)
				this.CurrentCell.CancelEdit();
			_toBeSaved = false;
			if(_deletedRecords != null && _deletedRecords.Count > 0)
				_deletedRecords.Clear();
		}
		#endregion

	
		protected override void OnTopRowChanged(GridRowColIndexChangedEventArgs e)
		{
			base.OnTopRowChanged (e);
			if(_dataSource != null && this.ViewLayout.LastVisibleRow > this.Model.RowCount - 30)
			{
				this.BeginUpdate();
				this.Binder.SuspendBinding();
				System.Reflection.ConstructorInfo oConstructor = _dataObjectType.GetConstructor(Type.EmptyTypes);
				this.AddItem(oConstructor.Invoke(null));
				this.Binder.ResumeBinding();
				this.Model.ResetVolatileData();
				this.EndUpdate();
				this.UpdateScrollBars();
			}
		}

		#region KeyDown	
		protected override void OnKeyDown(KeyEventArgs e)
		{	
			
			int curRow = this.CurrentCell.RowIndex;
			int curCol = this.CurrentCell.ColIndex;
			
			if(this.ReadOnly && (e.KeyCode == Keys.F2 || e.KeyCode == Keys.Delete))
			{
				e.Handled = true;
				return;
			}

			base.OnKeyDown(e);

			if(!this.ReadOnly && e.KeyCode == Keys.Insert && !e.Handled)
			{
				this.InsertRows();
				return;
			}

			if(!CurrentCell.IsEditing &&  curCol == CurrentCell.ColIndex && curRow == CurrentCell.RowIndex)
			{
				Keys modifierKeys = Control.ModifierKeys;
				bool extendSelection = (modifierKeys & Keys.Shift) != Keys.None;
				Keys keyCode = e.KeyCode & Keys.KeyCode;
				bool controlKeyDown = (e.Modifiers & Keys.Control) != Keys.None;
				bool shiftKeyDown = (e.Modifiers & Keys.Shift) != Keys.None;
				bool menuKeyDown = (e.Modifiers & Keys.Menu) != Keys.None;

				switch (keyCode)
				{
					case Keys.Tab:
						
						if (!controlKeyDown && !menuKeyDown)
						{
							if (shiftKeyDown && curRow > 1)
							{
								int i = 0;
								while(i < this.Model.ColCount - 1 && !CurrentCell.MoveTo(curRow - 1,this.Model.ColCount-i,GridSetCurrentCellOptions.ForceRefresh | GridSetCurrentCellOptions.ScrollInView )) 
									i++;
							}
							else 
							{							
								int i = 1;
								while(i <= this.Model.ColCount && !CurrentCell.MoveTo(curRow + 1,i,GridSetCurrentCellOptions.ScrollInView))
									i++;
							}

						}
						break;

					case Keys.Left:
						if (!menuKeyDown && !extendSelection)
						{
							if (curRow > 1)
							{
								int i = 0;
								while(i < this.Model.ColCount - 1 && !CurrentCell.MoveTo(curRow - 1,this.Model.ColCount-i,GridSetCurrentCellOptions.ForceRefresh | GridSetCurrentCellOptions.ScrollInView)) 
									i++;
							}
						}
						break;

					case Keys.Right:
						if (!menuKeyDown && !extendSelection)
						{
							int i = 1;
							while(i <= this.Model.ColCount && !CurrentCell.MoveTo(curRow + 1,i,GridSetCurrentCellOptions.ScrollInView))
								i++;
							
						}
						break;
					case Keys.Return:
						if (!menuKeyDown && !controlKeyDown && !shiftKeyDown)
						{
							int i = 1;
							while(i <= this.Model.ColCount && !CurrentCell.MoveTo(curRow + 1,i,GridSetCurrentCellOptions.ScrollInView))
								i++;
							
						}
						break;
					case Keys.Delete :
						if(!this.Binder.DirectSaveCellInfo)
							this.Binder.DirectSaveCellInfo = true;
						break;
				}			
			}
		}
		#endregion
		
		private void RaiseRecordAddedEvent(object record)
		{
			if(RecordAdded != null)
				RecordAdded(this,new RecordAddedEventArgs(record));
		}

		public void OnGridDataUpdated()
		{
			this._toBeSaved = true;
			this.Model.ResetVolatileData();
			this.RefreshRange(this.ViewLayout.VisibleCellsRange);
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public int DeletedRecordCount
		{
			get
			{
				if(_deletedRecords == null)
					return 0;
				else
					return _deletedRecords.Count;
			}
		}

		#region SUB TOTALS
		
		// SubTotal Fields;
		private string _subTotDescFld;
		private string _subTotFld;
		private string _subTotFlagFld;
		private ArrayList _subTotalRows;

		public void AddSubtotal(int rowNumber, int startRow, int endRow)
		{
		
			if(_subTotalRows == null)
				_subTotalRows = new ArrayList();

			string formulaString = "SUM(" + startRow + "," + endRow + ")";
			int colIndex = this.Binder.NameToColIndex(_subTotFld);
			//GesGridFormulaCellModel cellModel = (GesGridFormulaCellModel)this.Model.CellModels[FormulaCell];
			//This for loop is used when calculating a subtotal and that calculation includes a sub total row.
			for(int i = 0;i< _subTotalRows.Count ;i++)
			{
				int row = (int)_subTotalRows[i];
				if(row >= startRow && row <= endRow)
					formulaString += " - #" + row;
				if(row > endRow)
					break;
			}
			
			_subTotalRows.Add(rowNumber);
			_subTotalRows.Sort();
			

			//			_dataObjectType.InvokeMember(_subTotFld, BindingFlags.Default | BindingFlags.SetProperty ,
			//				null,_dataSource[rowNumber-1],new object[]{formulaString});

			
			//this.Model.ResetVolatileData();	
			//this.RefreshRange(this.ViewLayout.VisibleCellsRange);
			int oldStartRow	=	0;
			int oldEndRow	=	0;
			string dispText = string.Empty;
			for(int i = 0;i< _subTotalRows.Count ;i++)
			{
				int row = (int)_subTotalRows[i];
				if(row != rowNumber)
				{
					GetSubtotalStartEndRow(row ,out oldStartRow,out oldEndRow,out dispText);
					//This will get executed when subtotal(1,9) is there and u want to insert subtotal(1,3/4).
					if(oldStartRow <= rowNumber && oldEndRow >= rowNumber)
					{
						string oldFormula = (string)_dataObjectType.InvokeMember(_subTotFld,
							BindingFlags.Default | BindingFlags.GetProperty,null,_dataSource[row -1],null);
						oldFormula += " - #" + rowNumber;

						this[row,colIndex].Text = oldFormula;
						this[row,colIndex].FormulaTag = null;
						//					_dataObjectType.InvokeMember(_subTotFld, BindingFlags.Default | BindingFlags.SetProperty ,
						//						null,_dataSource[row-1],new object[]{oldFormula});
						//					cellModel.RecalculateRange(GridRangeInfo.Cell(row,colIndex));
						((GesGridRecordEntity)this._dataSource[row-1]).SetRowStateModified();
					}
				}
			}
			//this.Model.ResetVolatileData();
			this[rowNumber,colIndex].Text = formulaString;
			this[rowNumber,colIndex].FormulaTag = null;
			this.RefreshRange(this.ViewLayout.VisibleCellsRange);
			_toBeSaved = true;
		}
		public void DeleteSubtotal(int rowNumber)
		{
			if(_subTotalRows == null)
				return;
			//			for(int i = 0;i< _subTotalRows.Count ;i++)
			//			{
			//				if (rowNumber == _subTotalRows[i])
			//
			//			}
		
			//string formulaString = "SUM(" + startRow + "," + endRow + ")";
			int colIndex = this.Binder.NameToColIndex(_subTotFld);


			int oldStartRow = 0;
			int oldEndRow = 0;
			string dispText = string.Empty;
			for(int i = 0;i< _subTotalRows.Count ;i++)
			{
				int row = (int)_subTotalRows[i];
				if (row != rowNumber + 1)
				{
					GetSubtotalStartEndRow(row ,out oldStartRow,out oldEndRow,out dispText);
					//This will get executed when subtotal(1,9) is there and u want to insert subtotal(1,3/4).
					if(oldStartRow <= rowNumber + 1 && oldEndRow >= rowNumber + 1)
					{
						string oldFormula = (string)_dataObjectType.InvokeMember(_subTotFld,
							BindingFlags.Default | BindingFlags.GetProperty,null,_dataSource[row -1],null);
						string temp = "- #" + rowNumber;
						//MessageBox.Show(oldFormula.IndexOf(temp,0,oldFormula.Length).ToString());
						if (oldFormula.IndexOf(temp,0,oldFormula.Length)>0)
						{
							oldFormula = oldFormula.Remove(oldFormula.IndexOf(temp,0,oldFormula.Length)-1,5);
							//oldFormula += " + #" + rowNumber;

							this[row,colIndex].Text = oldFormula;
							this[row,colIndex].FormulaTag = null;
							//					_dataObjectType.InvokeMember(_subTotFld, BindingFlags.Default | BindingFlags.SetProperty ,
							//						null,_dataSource[row-1],new object[]{oldFormula});
							//					cellModel.RecalculateRange(GridRangeInfo.Cell(row,colIndex));
							((GesGridRecordEntity)this._dataSource[row-1]).SetRowStateModified();
						}
					}
				}
			}
			//this[rowNumber,colIndex].Text = formulaString;
			//this[rowNumber,colIndex].FormulaTag = null;
			this.RefreshRange(this.ViewLayout.VisibleCellsRange);
			_toBeSaved = true;
		}
		public void EditSubtotal(int rowNumber, int startRow, int endRow)
		{
			
			if(_subTotalRows == null)
				_subTotalRows = new ArrayList();

			string formulaString = "SUM(" + startRow + "," + endRow + ")";
			int colIndex = this.Binder.NameToColIndex(_subTotFld);
			//GesGridFormulaCellModel cellModel = (GesGridFormulaCellModel)this.Model.CellModels[FormulaCell];
			
			for(int i = 0;i< _subTotalRows.Count ;i++)
			{
				int row = (int)_subTotalRows[i];
				if(row >= startRow && row<= endRow)
					formulaString += " - #" + row;
				if(row > endRow)
					break;
			}

			this[rowNumber,colIndex].Text = formulaString;
			this[rowNumber,colIndex].FormulaTag = null;
			((GesGridRecordEntity)this._dataSource[rowNumber-1]).SetRowStateModified();
			this.RefreshRange(this.ViewLayout.VisibleCellsRange);
			_toBeSaved = true;
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public string SubtotalField
		{
			get{return _subTotFld;}
			set{_subTotFld = value;}
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public string SubtotalDescField
		{
			get{return _subTotDescFld;}
			set{_subTotDescFld = value;}
		}

		[Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public string SubtotalFlagField 
		{
			get{return _subTotFlagFld;}
			set{_subTotFlagFld = value;}
		}

		public void AddSubtotal(int rowNumber)
		{
			if(this._subTotalRows == null)
				_subTotalRows = new ArrayList();
			if(!this._subTotalRows.Contains(rowNumber))
			{
				_subTotalRows.Add(rowNumber);
				_subTotalRows.Sort();
			}
		}

		public void RefreshSubtotalRowNumbers(int rowNumber, bool isInsert, int count)
		{
			if(_subTotalRows == null || _subTotalRows.Count == 0)
				return;

			if(!isInsert)
			{
				if(_subTotalRows.Contains(rowNumber))
					_subTotalRows.Remove(rowNumber);
			}
			
			if(!isInsert)
				count *= -1;
			for(int i = 0;i < _subTotalRows.Count;i++)
			{
				int val = (int)_subTotalRows[i];
				if(val >= rowNumber)
				{
					val += count;
					_subTotalRows[i] = val;
				}
			}
		}

		public int GetLastSubtotalRow()
		{
			if(_subTotalRows == null || _subTotalRows.Count== 0)
				return 1;
			else
			{
				int rowIndex = this.CurrentCell.RowIndex;
				int lastRow = 1;
				for(int i =0 ; i< _subTotalRows.Count;i++)
				{
					int row = (int)_subTotalRows[i];
					if(row < rowIndex)
						lastRow = row + 1;
					else if(row > rowIndex)
						break;
				}
				return lastRow;
			}
		}

		public void GetSubtotalStartEndRow(int rowIndex, out int startRow, out int endRow, out string displayText)
		{
			int colIndex = this.NameToColIndex(_subTotFld);
			string formula = this[rowIndex,colIndex].Text;
			int pos = 4;
			string s= string.Empty;
			while(pos<formula.Length && char.IsDigit(formula[pos]))
			{
				s+=formula[pos];
				pos++;
			}
			startRow = int.Parse(s,System.Globalization.NumberStyles.Any);
			pos++;
			s = string.Empty;
			while(pos<formula.Length && char.IsDigit(formula[pos]))
			{
				s+=formula[pos];
				pos++;
			}
			endRow = int.Parse(s,System.Globalization.NumberStyles.Any);

			colIndex = this.NameToColIndex(_subTotDescFld);
			displayText = this[rowIndex,colIndex].Text;
		}
		#endregion

		#region UNDO Redo
		public void Undo()
		{
			GridModelCommandManager commandStack = this.Model.CommandStack;
			if(commandStack.UndoStack.Count == 0)
				return;

			bool endUpdate = false;
			try
			{
				while(commandStack.IsRecording && commandStack.UndoStack.Count > 0 && 
					commandStack.UndoStack.Peek() is GridSelectionStateCommand)
				{
					if (!endUpdate)
						this.Model.BeginUpdate(BeginUpdateOptions.InvalidateAndScroll);
					endUpdate = true;
					Execute(commandStack,GridCommandMode.Undo);
				}
				if(commandStack.UndoStack.Count > 0)
				{
					GridChangeCellsCommand  cmd  = commandStack.UndoStack.Peek() as GridChangeCellsCommand;
					if(cmd != null )
					{
						// Range is of the Format R8C12 - Row 8 Col 12
						string range = cmd.Description.Substring(CommandChangeCells.Length+1);
						
						int i =1;
						while(i< range.Length && char.IsDigit(range[i]))
							i++;
						int rowIndex = int.Parse(range.Substring(1,i-1));
						
						int j = ++i;
						while(i< range.Length && char.IsDigit(range[i]))
							i++;
						int colIndex = int.Parse(range.Substring(j,i-j));

						string oldValue = this[rowIndex,colIndex].Text;
						object oldrecord = ((ICloneable)this._dataSource[rowIndex-1]).Clone();
						
						commandStack.Undo();
									
						if(endUpdate)
							this.Model.EndUpdate();
						this.RecalculateRow(rowIndex);
						
					
						if(string.Compare(this[rowIndex,colIndex].Text,oldValue,true) == 0)
							this.Undo();
						else if(RecordEdited != null)
						{
							string fldName = this.Binder.InternalColumns[colIndex-1].MappingName;
							RecordEdited(this,new RecordEditedEventArgs(oldrecord,rowIndex-1,fldName));
						}
						this.RefreshRange(GridRangeInfo.Cells(rowIndex ,0 ,rowIndex,this.ViewLayout.LastVisibleCol));
					}
					else
						commandStack.Undo();
				}
			}
			finally
			{
				if(endUpdate)
					this.Model.EndUpdate();
			}
		}

		public void Redo() 
		{
			GridModelCommandManager commandStack = this.Model.CommandStack;
			if(commandStack.RedoStack.Count == 0)
				return;
			bool endUpdate = false;
			try
			{
				if (commandStack.IsRecording)
				{
					if (commandStack.RedoStack.Count > 0 && commandStack.RedoStack.Peek() is GridSelectionStateCommand)						
					{
						if(!endUpdate)
						{
							this.Model.BeginUpdate(BeginUpdateOptions.InvalidateAndScroll);
							endUpdate = true;
						}
						while (commandStack.RedoStack.Count > 0 && commandStack.RedoStack.Peek() is GridSelectionStateCommand)
							Execute(commandStack, GridCommandMode.Redo);
					}
				}
				if(commandStack.RedoStack.Count > 0)
				{
					GridChangeCellsCommand  cmd  = commandStack.RedoStack.Peek() as GridChangeCellsCommand;
					if(cmd != null )
					{
						// Range is of the Format R8C12 - Row 8 Col 12
						string range = cmd.Description.Substring(CommandChangeCells.Length+1);
						int i =1;
						while(i< range.Length && char.IsDigit(range[i]))
							i++;
						int rowIndex = int.Parse(range.Substring(1,i-1));
						
						int j = ++i;
						while(i< range.Length && char.IsDigit(range[i]))
							i++;
						int colIndex = int.Parse(range.Substring(j,i-j));

						string oldValue = this[rowIndex,colIndex].Text;
						object oldRecord = ((ICloneable)this._dataSource[rowIndex-1]).Clone();
						
						commandStack.Redo();
					
						if(RecordEdited != null)
						{
							string fldName = this.Binder.InternalColumns[colIndex-1].MappingName;
							RecordEdited(this,new RecordEditedEventArgs(oldRecord,rowIndex-1,fldName));
						}
						if(endUpdate)
							this.Model.EndUpdate();
						this.RecalculateRow(rowIndex);
						this.RefreshRange(GridRangeInfo.Cells(rowIndex ,0 ,rowIndex,this.ViewLayout.LastVisibleCol));
						if(string.Compare(this[rowIndex,colIndex].Text,oldValue,true) == 0)
							this.Redo();
					}
					else
						commandStack.Redo();
				}
			}
			finally
			{
				if (endUpdate)
					Model.EndUpdate();
			}
		}

		void Execute(GridModelCommandManager commandStack, GridCommandMode mode)
		{
			if (commandStack.InTransaction)
				throw new InvalidOperationException("Can't perform Undo or Redo during a transaction ") ;
			
			GridCommandMode savedMode = commandStack.Mode;
			commandStack.Mode  = mode;
			SyncfusionCommand cmd = null;
			if(mode == GridCommandMode.Undo)
				cmd = commandStack.UndoStack.Pop() as  SyncfusionCommand; //stack.Pop() as SyncfusionCommand;
			else if( mode == GridCommandMode.Redo)
				cmd = commandStack.RedoStack.Pop() as SyncfusionCommand;
			
			try
			{
				if (cmd != null)
				{
					cmd.Execute();
					//if ((mode == GridCommandMode.Redo ? cmd : Peek()) is GridSelectionStateCommand)
					//	Execute(mode);
				}
			}
			catch (Exception ex)
			{
				if (!ExceptionManager.RaiseExceptionCatched(this, ex))
					throw ex;
			}  
			commandStack.Mode = savedMode;
		}


		#endregion

		#region Column Settings
		private void Model_ColWidthsChanged(object sender, GridRowColSizeChangedEventArgs e)
		{
			if(e.Success && !_inSetColWidths)
			{
				int index = 0;
				for(int i = e.From-1; i< e.To;i++)
				{
					string fldName = this.Binder.InternalColumns[i].MappingName;
					GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[(int)_fieldIndexes[fldName]];
					if(colStyle.Visible)
						colStyle.CurrentWidth = (short)this.Model.ColWidths[fldName];
					index++;
				}
				_colSettingsChanged = true;
			}
		}

		private void Model_ColsMoved(object sender, GridRangeMovedEventArgs e)
		{
			if(e.Success)
			{
				for(int i=0; i< _columnInfo.Count;i++)
				{
					GesColumnStyle colStyle = (GesColumnStyle)_columnInfo[i];
					colStyle.ColOrder = (byte)this.Binder.NameToColIndex(colStyle.FieldName);
				}
				_colSettingsChanged = true;
			}
		}

		#endregion

		public GridRangeInfoList GetSelectedRows()
		{
			GridRangeInfoList rangeList;
			// If the entire table is selected return all the rows
			if(this.Selections.Ranges.ActiveRange.IsTable)
			{
				rangeList = new GridRangeInfoList();
				rangeList.Add(GridRangeInfo.Rows(1,this.Model.RowCount));
			}
			else
				rangeList = this.Selections.GetSelectedRows(true, false);

			return rangeList;
		}

		// Overrided to avoid the celldouble click incase of notecell
		protected override void OnCellDoubleClick(GridCellClickEventArgs e)
		{
			GridStyleInfo style = this.Model[e.RowIndex, e.ColIndex];
			if(style.CellType != NoteCell)
				base.OnCellDoubleClick (e);
		}
		protected override bool ProcessKeyMessage(ref Message m)
		{
			Keys keyCode = (Keys)((int)m.WParam) & Keys.KeyCode;
			switch (keyCode)
			{
				case Keys.Home:
					if (Control.ModifierKeys == Keys.Control)
						this.CurrentCell.MoveTo(1,this.CurrentCell.ColIndex); 
					else
						this.CurrentCell.MoveTo(this.CurrentCell.RowIndex,1);
					return true;
					break;
				case Keys.End:
					if (Control.ModifierKeys == Keys.Control)
					{
						this.CurrentCell.MoveTo(this.GridCount,this.CurrentCell.ColIndex);
						return true;
					}
					else
					{
						this.CurrentCell.MoveTo(this.CurrentCell.RowIndex,this.Model.ColCount-1);
						return true;
					}
					break;
			}
			return base.ProcessKeyMessage(ref m);
		}
		private int GridCount
		{
			set{_gridCount = value;}
			get{return _gridCount;}
		}
	}
}
