using System;
using Syncfusion.Windows.Forms.Grid;
using System.Windows.Forms;
using System.Drawing;
using System.ComponentModel;
using System.Security;
using System.Security.Permissions;
using System.Runtime.InteropServices;

namespace GES.GridCellModels
{
	/// <summary>
	/// Summary description for GESRichTextBox.
	/// </summary>
	public class GESRichTextBox : RichTextBox
	{
		private Size textSize = new Size();
		private InPlaceRichTextCellRenderer parent;
		private bool allowNavigation;
		//bool ignoreWmPaste = false;

		public const int WM_PASTE = 770; // 0x0302 
		public const int WM_KEYUP = 257; // 0x0101 
		
		[DllImport("user32", CharSet=CharSet.Auto)]
		extern public static IntPtr SendMessage(IntPtr hWnd, int Msg, int wParam, int lParam)  ;
		/// <summary>
		/// Returns the associated cell renderer for the text box.
		/// </summary>
		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public GridCellRendererBase ParentCell
		{
			get
			{
				return this.parent;
			}
		}

		public GESRichTextBox(InPlaceRichTextCellRenderer parent)
		{
			this.parent = parent;
			this.BorderStyle = System.Windows.Forms.BorderStyle.None;
			this.CausesValidation = false;
			this.SetStyle(ControlStyles.Selectable, false);
			this.SetStyle(ControlStyles.StandardClick|ControlStyles.StandardDoubleClick, true);
			this.AutoWordSelection = false;
			this.AllowDrop = true;
			this.MouseDown+=new MouseEventHandler(GESRichTextBox_MouseDown);
			this.DragEnter += new System.Windows.Forms.DragEventHandler(GESRichTextBox_DragEnter);
			this.DragDrop += new System.Windows.Forms.DragEventHandler(GESRichTextBox_DragDrop);
		}

		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged (e);
			
		}

		public bool AllowNavigation
		{
			get{return allowNavigation;}
			set{allowNavigation = value;}
		}
		/// <summary>
		/// Gets / sets if the cell supports floating over another cell.
		/// </summary>
		
		protected override void OnContentsResized(ContentsResizedEventArgs e)
		{
			base.OnContentsResized (e);
			if (parent != null)
			{
				textSize.Height = e.NewRectangle.Height;//this.GetLineCount();//
				textSize.Width = e.NewRectangle.Width;

				base.OnContentsResized(e);

				if (!(Visible && Focused ))
					return;

				Size size = this.parent.Grid.Model.AddBorders(textSize, this.parent.StyleInfo);

				GridRangeInfo rangeCell = GridRangeInfo.Cell(this.parent.RowIndex, this.parent.ColIndex);

				this.parent.Grid.Model.FloatingCells.DelayFloatCells(rangeCell);
				
				this.parent.Grid.Model.GetSpannedRangeInfo(this.parent.RowIndex, this.parent.ColIndex, out rangeCell);
				
				if (textSize.Height > ClientRectangle.Height)
				{
					this.parent.Grid.Model.RowHeights.ResizeToFit(rangeCell, GridResizeToFitOptions.ResizeCoveredCells);
					Update();
					LineScroll(-1);
					LineScroll(-1);
				}

				if (textSize.Width > ClientRectangle.Width)
				{
					this.parent.Grid.Model.ColWidths.ResizeToFit(rangeCell, GridResizeToFitOptions.ResizeCoveredCells);
					//        .SetColWidth(this.parent.currentColIndex, this.parent.currentColIndex, size.Width, null);
				}
				if(textSize.Height < ClientRectangle.Height)
				{
					this.parent.Grid.Model.RowHeights.ResizeToFit(rangeCell, GridResizeToFitOptions.ResizeCoveredCells);
				}
			}
		}

		[SecurityPermission(SecurityAction.Demand, UnmanagedCode=true)]
		public void LineScroll(int count)
		{
			SendMessage(Handle, 182/*EM_LINESCROLL*/, 0, count);
		}

		/// <override/>
		[SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode=true)]
		protected override bool ProcessKeyMessage(ref Message m)
		{
			//			if (m.Msg == WM_KEYUP)
			//				ignoreWmPaste = false;

			InPlaceRichTextCellRenderer  tbr = ParentCell as InPlaceRichTextCellRenderer ;
			if (tbr == null)
				return base.ProcessKeyMessage(ref m);
			
			Rectangle intersect = Rectangle.Intersect(this.Bounds, this.parent.Grid.GridBounds);

			Keys keyCode = (Keys)((int)m.WParam) & Keys.KeyCode;
			if (m.Msg == 0x102/*WM_CHAR*/)
			{
				if (parent.Grid.WantEnterKey && keyCode == Keys.Enter
					|| parent.Grid.WantEscapeKey && keyCode == Keys.Escape)
					return true;

				if (intersect != this.Bounds)
					this.parent.Grid.CurrentCell.ScrollInView(GridScrollCurrentCellReason.KeyPress);
				bool b = ParentCell.RaiseProcessKeyEventArgs(ref m)
					|| this.ProcessKeyEventArgs(ref m);
				//return this.ProcessKeyEventArgs(ref m);
				return b;
			}

			bool bCtl = (Control.ModifierKeys & Keys.Control) != Keys.None;
			bool bAlt = (Control.ModifierKeys & Keys.Alt) != Keys.None;
			bool bShift = (Control.ModifierKeys & Keys.Shift) != Keys.None;

			bool forwardParent = false;
			bool scrollInView = false;
			bool callBase = true;

			Keys kc = keyCode;
			if (RightToLeft == RightToLeft.Yes)
			{
				switch (kc)
				{
					case Keys.Left:
						kc = Keys.Right;
						break;
					case Keys.Right:
						kc = Keys.Left;
						break;
				}
			}
			switch (kc)
			{
				case Keys.Home:
				case Keys.End:
					if (bCtl)
						forwardParent = true;
					else
						scrollInView = true;
					callBase = false;
					break;

				case Keys.Left:
					if (bCtl || allowNavigation && this.SelectionStart <= 0 && (this.SelectionLength == 0 || !bShift))
						forwardParent = true;
					else
						scrollInView = true;
					callBase = false;
					break;
				case Keys.Right:
					if (bCtl || allowNavigation && this.SelectionStart + this.SelectionLength >= this.Text.Length)
						forwardParent = true;
					else
						scrollInView = true;
					callBase = false;
					break;
				case Keys.Up:
					if (bCtl || allowNavigation && this.GetLineFromCharIndex(SelectionStart) <= 0)
						forwardParent = true;
					else
						scrollInView = true;
					callBase = false;
					break;
				case Keys.Down:
					if (bCtl || allowNavigation && this.GetLineFromCharIndex(this.SelectionStart + this.SelectionLength) >= this.GetLineFromCharIndex(this.TextLength))
					{
						callBase = !bAlt;
						forwardParent = !bAlt;
					}
					else
					{
						scrollInView = true;
						callBase = false;
					}
					break;
				case Keys.PageDown:
				case Keys.PageUp:
				case Keys.Tab:
				case Keys.F2:
					forwardParent = true;
					callBase = false;
					break;

				case Keys.Escape:
					if (parent.Grid.WantEscapeKey)
					{
						forwardParent = true;
						scrollInView = true;
						callBase = false;
					}
					break;

				case Keys.Enter:
					// FYI - In the following line, we define the behavior when the
					// user presses <CTRL>-<Enter>. Right now, the Enter will insert a newline
					// if <CTRL> is pressed.
					if (parent.Grid.WantEnterKey)
					{
						if (!this.parent.StyleInfo.AllowEnter && !bCtl)
							forwardParent = true;
						scrollInView = true;
						callBase = false;
					}
					break;

				case Keys.V:
					scrollInView = bCtl;
					forwardParent =  bCtl && this.SelectionLength == 0 || bAlt;
					callBase = false;
					break;

				case Keys.X:
				case Keys.C:
					scrollInView = bCtl;
					forwardParent = false;
					callBase = true;
					break;

				case Keys.Insert:
					scrollInView = true;
					forwardParent = (bCtl || bShift) && this.SelectionLength == 0 || bAlt; // || bShift;
					callBase = false;
					break;

				case Keys.Delete:

					scrollInView = true;
					forwardParent = (bCtl || bAlt || bShift) && this.SelectionLength == 0;
					callBase = false;
					break;

				case Keys.F4:
					if (bCtl || bAlt)
					{
						return base.ProcessKeyMessage(ref m);
					}
					break;

				case Keys.ShiftKey:
				case Keys.ControlKey:
					forwardParent = true;
					callBase = false;
					break;

				default:
					
					forwardParent = bCtl || bAlt;
					break;
			}

			// Give programmers a chance to modify the default behavior of ProcessKeyMessage
			// without subclassing this control.
			GridCurrentCellControlKeyMessageEventArgs e = new GridCurrentCellControlKeyMessageEventArgs(this, m, scrollInView, forwardParent, callBase, true);
			this.parent.Grid.RaiseCurrentCellControlKeyMessage(e);
			if (e.Handled)
				return e.Result;
			scrollInView = e.ScrollInView;
			forwardParent = e.CallProcessKeyPreview;
			callBase = e.CallBaseProcessKeyMessage;


			if (scrollInView)
			{
				if (intersect != this.Bounds)
					this.parent.Grid.CurrentCell.ScrollInView(GridScrollCurrentCellReason.KeyPress);
			}

			if (forwardParent)
			{
				// This will trigger Grid.OnKeyDown and Grid.OnKeyUp events. CurrentCellKeyDown and CurrentCellKeyPress events might also
				// be triggered from Grids ProcessKeyPreview method.
				this.ProcessKeyPreview(ref m);
				return true;
			}
			else
			{
				// Call to ParentCell.ProcessKeyEvenspecialtArgs will trigger CurrentCellKeyDown and CurrentCellKeyPress events but no
				// Grid.OnKeyDown and Grid.OnKeyUp events.
				if (callBase)
					return base.ProcessKeyMessage(ref m);
				else
				{
					// Call to ParentCell.ProcessKeyEventArgs will trigger CurrentCellKeyDown and CurrentCellKeyPress events but no
					// Grid.OnKeyDown and Grid.OnKeyUp events.
					bool b = ParentCell.RaiseProcessKeyEventArgs(ref m)
						|| this.ProcessKeyEventArgs(ref m);

					return b;
				}
			}
		}

		private void GESRichTextBox_MouseDown(object sender, MouseEventArgs e)
		{
			DragDropEffects DragDropResult = new DragDropEffects();
			if (e.Button == MouseButtons.Right)
			{
				DragDropResult = this.DoDragDrop(this.SelectedText,DragDropEffects.Move | DragDropEffects.Copy);
					
				//				if (DragDropResult == DragDropEffects.Move)
				//					listBox1.Items.RemoveAt(textBox1.SelectedText);
			}
		}
		private void GESRichTextBox_DragEnter(object sender, System.Windows.Forms.DragEventArgs e)
		{
			if (e.KeyState==9)
				e.Effect = DragDropEffects.Copy;
			else
				e.Effect = DragDropEffects.Move;
		}
		private void GESRichTextBox_DragDrop(object sender, System.Windows.Forms.DragEventArgs e)
		{
			
			MessageBox.Show(this.PointToClient(new Point(e.X,e.Y)).ToString());
		
			e.Data.GetData(DataFormats.Text);
			
		}
	}
}
